/**
 * Copyright (c) 2014 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather.db;

import java.io.File;
import java.io.FileOutputStream;

import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;

import java.util.Date;

import weather.Utils;


/**
 * The manager for locking of the database for safe concurrent access.
 * This is a static class since the database files (and in particular the lock file)
 * are a shared resource.
 */
public class LockManager
{
	/**
	 * The lock file name contained within the database directory.
	 */
	public static final String LOCK_FILE_NAME = "DATABASE_LOCK_FILE";

	/**
	 * The timeout in milliseconds while waiting for the lock before giving up.
	 */
	public static final long LOCK_TIMEOUT = 1000 * 30;

	/**
	 * The time in milliseconds while waiting for the lock when we warn the user.
	 */
	public static final long LOCK_WARN_TIME = 1000 * 2;

	/**
	 * The time in milliseconds to sleep while waiting for the lock.
	 */
	public static final long LOCK_SLEEP_TIME = 250;

	/**
	 * The data objects used to perform the locking.
	 * The lock is owned by us if the mFileLock is non-null.
	 */
	private static File mLockFile;
	private static FileOutputStream mOutputStream;
	private static FileChannel mFileChannel;
	private static FileLock mFileLock;


	/**
	 * Return a File object for the database lock file.
	 * The lock file itself may or may not exist.
	 * Returns null if there is no database directory defined.
	 */
	public static File getDatabaseLockFile()
	{
		if (mLockFile != null)
			return mLockFile;

		File databaseDirectory = DBUtils.getDatabaseDirectory();

		if (databaseDirectory == null)
			return null;

		mLockFile = new File(databaseDirectory, LOCK_FILE_NAME);

		return mLockFile;
	}


	/**
	 * Get the lock on the database if we can, waiting if necessary,
	 * Returns immediately if the database is already locked by us
	 * (but note that the locking is not reference counted).
	 * Returns true on success.  If the database directory is not
	 * configured then this appears to succeed.
	 * @return
	 */
	public static boolean getLock()
	{
		if (mFileLock != null)
			return true;

		File lockFile = getDatabaseLockFile();

		if (lockFile == null)
			return true;

		try
		{
			//
			// Open the lock file and its channel for writing if necessary.
			//
			if (mOutputStream == null)
				mOutputStream = new FileOutputStream(lockFile);

			if (mFileChannel == null)
				mFileChannel = mOutputStream.getChannel();

			//
			// Now try to lock the file, timing out if we wait too long.
			//
			return getLockWithTimeout();
		}
		catch (Exception e)
		{
			Utils.logError("Cannot acquire lock file " + lockFile.getPath(), e);

			return false;
		}
	}


	/**
	 * Try to lock the opened file while handling a timeout.
	 * Returns true if the lock was successfully obtained or false
	 * if we timed out waiting for the lock.
	 */
	private static boolean getLockWithTimeout() throws Exception
	{
		//
		// First try to get the lock right away which usually works.
		//
		mFileLock = mFileChannel.tryLock();

		if (mFileLock != null)
			return true;

		//
		// We have to wait for the lock.
		// Loop trying to get it until the timeout expires.
		// Give a warning message if we wait for a little while.
		//
		long startTime =  new Date().getTime();
		boolean isWarned = false;

		while (true)
		{
			Thread.sleep(LOCK_SLEEP_TIME);

			mFileLock = mFileChannel.tryLock();

			if (mFileLock != null)
				return true;

			long thisTime = new Date().getTime();

			if (thisTime > startTime + LOCK_TIMEOUT)
			{
				Utils.logError("Timeout waiting for database lock");

				return false;
			}

			if (!isWarned && (thisTime > startTime  + LOCK_WARN_TIME))
			{
				Utils.logError("Waiting for database lock ...");

				isWarned = true;
			}
		}
	}


	/**
	 * Free the lock file if necessary.
	 * Nothing is done if the database was not locked by us.
	 * The lock file is not closed here.
	 * This catches any exceptions.
	 */
	public static void freeLock()
	{
		if (mFileLock == null)
			return;

		try
		{
			mFileLock.release();
		}
		catch (Exception e)
		{
			Utils.logError("Exception releasing file lock", e);
		}
		finally
		{
			mFileLock = null;
		}
	}


	/**
	 * Free and close the lock file if necessary.
	 * This is called when cleaning up before exiting.
	 */
	public static void closeLock()
	{
		freeLock();

		Utils.safeClose(mOutputStream);

		mOutputStream = null;
		mFileChannel = null;
	}
}
