/**
 * Copyright (c) 2014 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather.db;

import java.util.HashMap;

import java.io.File;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;

import weather.Utils;


/**
 * A class to save the URLs and site names about the weather sites that
 * have been collected by the program.  (This could almost be obtained
 * using the site definitions in the configuration file, but the site names
 * there aren't the real ones obtained from the reports.)  All of the
 * information is saved in one file within the database directory.  This
 * class also acts as a cache of the information.
 */
public class SiteURLFile
{
	/**
	 * A mapping from the URLs to the site names.
	 */
	private HashMap<String, String> mURLMap;


	/**
	 * Return the site name associated with the specified URL if any.
	 * This loads the mappings from the database file if necessary. 
	 * Returns null if the site name is not known.
	 * @param URL
	 */
	public String getSiteName(String URL)
	{
		if (mURLMap == null)
			loadURLs();

		return mURLMap.get(URL);
	}


	/**
	 * Set the site name mapping for the specified URL.
	 * This writes any new information to the database file.
	 * Nothing is done if the site name is empty or the mapping
	 * already exists
	 * @param URL
	 * @param siteName
	 */
	public void setSiteName(String URL, String siteName)
	{
		if ((siteName == null) || siteName.isEmpty())
			return;

		//
		// If the mapping for the URL already exists and is still
		// the same as this one then do nothing.
		//
		String oldSiteName = getSiteName(URL);

		if (siteName.equals(oldSiteName))
			return;

		//
		// If the mapping existed but is different than before
		// then complain about it but proceed anyway.
		//
		if (oldSiteName != null)
		{
			Utils.logError("Site name for " + URL +
				" changed from " + oldSiteName + " to " + siteName);
		}

		//
		// Update the mapping and save them back to the file.
		//
		mURLMap.put(URL, siteName);

		saveURLs();
	}


	/**
	 * Load the URL and site name mappings from the database
	 * file into the hash map if possible.  An error is logged if the
	 * file exists but cannot be read properly.  The mapping is
	 * created empty if no mappings were found.
	 */
	public void loadURLs()
	{
		//
		// Create the mapping so it will always exist even if
		// there is an error reading the file.
		//
		mURLMap = new HashMap<String, String>();

		//
		// Get the handle on the database file.
		// Do nothing if the file doesn't exist yet.
		//
		File file = DBUtils.getSiteURLFile();

		if ((file == null) || !file.exists())
			return;

		//
		// The file does exist, so read its lines.
		// Since we are the only ones to generate this file we
		// assume that its contents will parse correctly.
		//
		FileReader fileReader = null;

		try
		{
			fileReader = new FileReader(file);
			BufferedReader bufferedReader = new BufferedReader(fileReader);

			while (true)
			{
				String line = bufferedReader.readLine();

				if (line == null)
					break;

				//
				// Break up the line into the URL and site name which
				// are separated by a single space, and add the result
				// into the map.
				//
				int index = line.indexOf(' ');

				String URL = line.substring(0, index);
				String siteName = line.substring(index + 1);

				mURLMap.put(URL, siteName);
			}

			fileReader.close();
			fileReader = null;
		}
		catch (Exception e)
		{
			Utils.logError("Exception reading " + file.getPath(), e);
		}
		finally
		{
			Utils.safeClose(fileReader);
		}
	}


	/**
	 * Save the URL and site name mappings to the database file.
	 * Nothing is done if the mappings are empty.
	 */
	public void saveURLs()
	{
		if ((mURLMap == null) || mURLMap.isEmpty())
			return;

		File file = DBUtils.getSiteURLFile();

		if (file == null)
			return;

		FileWriter writer = null;

		try
		{
			writer = new FileWriter(file);

			//
			// Write each of the mappings to the file as lines
			// having the format "URL siteName"
			//
			for (String URL : mURLMap.keySet())
			{
				String siteName = mURLMap.get(URL);

				writer.append(URL + ' ' + siteName + '\n');
			}

			writer.close();
			writer = null;
		}
		catch (Exception e)
		{
			Utils.logError("Failed to write " + file.getPath(), e);
		}
		finally
		{
			Utils.safeClose(writer);
		}
	}
}
