/**
 * Copyright (c) 2016 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather.gui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Stroke;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.util.Date;
import java.util.TimeZone;

import javax.swing.JFrame;
import javax.swing.JPanel;

import weather.Configuration;
import weather.Controller;
import weather.DateTime;
import weather.Observation;
import weather.ObservationList;
import weather.Report;
import weather.PlotParameters;
import weather.SunriseSunset;
import weather.Utils;
import weather.selector.Selector;


/**
 * A panel to plot observations.
 * This can show either one or two plots at the same time.
 * If two plots are shown and are compatible (i.e., they have the same units)
 * than the labels and grid lines are meaningful for both plots.  Otherwise the
 * second plot is shown without any labels or grid.
 */
public class PlottingPanel extends JPanel
{
	private static final long serialVersionUID = 1L;

	private int mHashedLineSize = 4;
	private int mHashedLineInterval = 10;
	private int mSideInset = 40;
	private int mTopInset = 10;
	private int mBottomInset = 50;

	private Controller mController;

	private BufferedImage mImage;
	private Graphics2D mGraphics;
	private int mGridWidth;
	private int mGridHeight;

	private int mPointRadius;
	private int mPointShrinkageDays;
	private int mDayTimeDays;
	private int mWindDirectionRadius;
	private Color mBackgroundColor;
	private Color mNightTimeBackgroundColor;
	private Color mDayTimeBackgroundColor;
	private Color mMajorHorizontalLineColor;
	private Color mMinorHorizontalLineColor;
	private Color mMajorVerticalLineColor;
	private Color mMinorVerticalLineColor;
	private Color mMarkerLineColor;
	private Color mLabelColor;
	private Color mDayOfWeekColor;

	private float [] mDashes = { 1f, 1f, 1f };

	private BasicStroke mDashesStroke = new BasicStroke(1,
		BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 1.0f, mDashes, 2f);

	private Stroke mDefaultStroke;
	private AffineTransform mDefaultTransform;

	/**
	 * The most recent set of plotting parameters that has been set.
	 */
	private PlotParameters mPlotParameters;

	//
	// The minimum and maximum UTC times for the plot.
	// This can be before or after any of the observations due to rounding.
	//
	private DateTime mMaximumTime;
	private DateTime mMinimumTime;

	/**
	 * The table of the sunrise and sunset times for the current observation list.
	 * The length of the table is the number of "noons" that the times use.
	 */
	private SunriseSunsetInfo[] mSunriseSunsetTable;

	/**
	 * The table of local times to be plotted.
	 * This is used for determining the major and minor vertical lines
	 * and also for the time labels at the bottom.  The table is indexed
	 * by 1/2 hour intervals from the minimum to maximum times.
	 */
	private DateTime[] mTimeTable;

	/**
	 * The report which this plot uses.
	 * The observation list which is plotted can contain more or less
	 * data than what is contained in the report.
	 */
	private Report mReport;

	/**
	 * The information about the first plot (which is always the primary one).
	 * This plot determines the type of value which is shown by the vertical
	 * labels and determines the horizontal grid lines for the major and
	 * minor intervals.
	 */
	private final PlotInfo mPlotInfo1 = new PlotInfo();

	/**
	 * The information about the second plot (which is optional).
	 * If the value types are compatible with the first plot then this
	 * plot shares the labels and grid lines with the first plot.
	 * If the value types are incompatible then the second plot will
	 * be shown but without any grid lines or labels of its own.
	 */
	private final PlotInfo mPlotInfo2 = new PlotInfo();

	private boolean mIsConsistent;
	private JFrame mOwnerFrame;

	private int mTimeMajorInterval;
	private int mTimeMinorInterval;

	private final TimeZone mGMTTimezone = TimeZone.getTimeZone("GMT");


	/**
	 * Constructor which accepts the owning frame and our controller.
	 * @param ownerFrame
	 * @param controller
	 */
	public PlottingPanel(JFrame ownerFrame, Controller controller)
	{
		mOwnerFrame = ownerFrame;
		mController = controller;

		mController.setPlottingPanel(this);

		Configuration config = Configuration.getInstance();

		setPreferredSize(new Dimension(config.getPlotWidth(), config.getPlotHeight()));

		mPointRadius = config.getPointRadius();
		mPointShrinkageDays = config.getPointShrinkageDays();
		mWindDirectionRadius = config.getWindDirectionRadius();
		mDayTimeDays = config.getDayTimeDays();

		getColors();

		setBackground(mBackgroundColor);

		setEnabled(true);
		setFocusable(true);

		//
		// Add event handlers for the keyboard and mouse.
		//
		addKeyListener(new KeyAdapter()
			{
				@Override
				public void keyPressed(KeyEvent e)
				{
					handleKeyPressed(e);
				}

				@Override
				public void keyTyped(KeyEvent e)
				{
					handleKeyTyped(e);
				}
			});

		addMouseListener(new MouseAdapter()
			{
				@Override
				public void mouseClicked(MouseEvent e)
				{
					handleMouseClicked(e);
				}
			});
	}


	/**
	 * Update the plot using the specified set of parameters.
	 * The isFullUpdate flag is set to indicate that the time interval or site
	 * is changed and so any cached time information must be recomputed.
	 */
	public void updatePlot(PlotParameters plotParameters, boolean isFullUpdate)
	{
		mPlotParameters = plotParameters;

		//
		// Copy some of the parameters for convenient access.
		// In particular, copy some values into the PlotInfo objects.
		//
		mReport = plotParameters.mReport;

		mPlotInfo1.mFullObservationList = plotParameters.mFullObservationList;
		mPlotInfo1.mObservationList = plotParameters.mObservationList1;
		mPlotInfo1.mSelector = plotParameters.mSelector1;

		mPlotInfo2.mFullObservationList = plotParameters.mFullObservationList;
		mPlotInfo2.mObservationList = plotParameters.mObservationList2;
		mPlotInfo2.mSelector = plotParameters.mSelector2;

		mMinimumTime = plotParameters.mMinimumTime;
		mMaximumTime = plotParameters.mMaximumTime;
		mTimeMajorInterval = plotParameters.mTimeMajorInterval;
		mTimeMinorInterval = plotParameters.mTimeMinorInterval;

		mIsConsistent = plotParameters.mFullObservationList.isConsistent();

		//
		// If this is a full update then recalculate the time information
		// and the sunrise/sunset tables.  Otherwise we assume that the
		// previously calculated values are still valid.
		//
		if (isFullUpdate)
		{
			//
			// Calculate the times for the vertical lines and bottom
			// bottom labels to use.
			//
			calculateTimeValues();

			//
			// Calculate the sunrise and sunset times for all the
			// days in the time interval.
			//
			calculateSunriseSunsetTimes();
		}

		//
		// Now cause the plot to be redrawn.
		// First invalidate the buffered image so that it will be recreated.
		//
		mImage = null;

		setTitle();
		repaint();
	}


	/**
	 * Set the title for the owning Frame to display the report info
	 * and the selector info.
	 */
	private void setTitle()
	{
		if ((mReport == null) || !mPlotInfo1.isValid())
			return;

		String title = mReport.getSiteName() + ": " +
			mReport.getRefreshMessage() + " - " +
			mPlotInfo1.mSelector.getName();

		//
		// Add the information about the second plot if any.
		//
		if (mPlotInfo2.isValid() && (mPlotInfo2.mSelector != mPlotInfo1.mSelector))
		{
			title += " + " + mPlotInfo2.mSelector.getName();

			if (mPlotInfo1.mSelector.getValueType() != mPlotInfo2.mSelector.getValueType())
				title += " (NO COORDINATES)";
		}

		//
		// Add information about grouping if any.
		//
		if (mPlotParameters.mGroupHours > 0)
		{
			String text;

			switch (mPlotParameters.mGroupHours)
			{
				case 12: text = "Half Days"; break;
				case 24: text = "Days"; break;
				case 7 * 24: text = "7 Days"; break;
				case 30 * 24: text = "30 Days"; break;
				default: text = mPlotParameters.mGroupHours + " hours"; break;
			}

			title += " (Min/Max grouping: " + text + ")";
		}

		mOwnerFrame.setTitle(title);
	}


	/**
	 * Draw this component.
	 * This is called on a refresh or due to our calling repaint.
	 * We use a buffered image because drawing into a memory buffer can
	 * be lots faster than drawing directly to the screen.
	 */
	public void paintComponent(Graphics graphics)
	{
		//
		// Set the screen to blank until we are finished drawing.
		//
		graphics.setColor(mBackgroundColor);
		graphics.fillRect(0, 0, getWidth(), getHeight());

		if ((mReport == null) || !mPlotInfo1.isValid())
			return;

		//
		// Make sure that the first plot has at least one value to be plotted.
		// Check the full observation list instead of the current one so that
		// gaps in the observations can be skipped over.
		//
		if (!mPlotInfo1.mSelector.hasValues(mPlotInfo1.mFullObservationList))
			return;

		//
		// Something exists to be drawn.
		// Try to calculate the size of the plotting grid.
		//
		mSideInset = mPlotInfo1.mSelector.getSideInset();

		mGridWidth = getWidth() - mSideInset * 2;
		mGridHeight = getHeight() - mTopInset - mBottomInset;

		if ((mGridWidth <= 0) || (mGridHeight <= 0))
			return;

		//
		// See if still have a buffered image of the correct size.
		// If not then invalidate it so we will create a new one.
		//
		if ((mImage != null) &&
			((mImage.getWidth() != getWidth()) ||
				(mImage.getHeight() != getHeight())))
		{
			mImage = null;
		}

		//
		// If we have a previous buffered image then just copy that to
		// the screen.
		//
		if (mImage != null)
		{
			graphics.drawImage(mImage, 0, 0, null);

			return;
		}

		//
		// Create a new buffered image for the new plot.
		// Then get the graphics object from it so that we can draw into it.
		//
		mImage = new BufferedImage(getWidth(), getHeight(),
				BufferedImage.TYPE_3BYTE_BGR);

		mGraphics = (Graphics2D) mImage.getGraphics();
		mDefaultStroke = mGraphics.getStroke();
		mDefaultTransform = mGraphics.getTransform();

		mGraphics.setColor(mBackgroundColor);
		mGraphics.fillRect(0, 0, getWidth(), getHeight());

		//
		// Calculate the minimum and maximum values to be plotted
		// and the major and minor intervals for that range.
		//
		calculateValuesRange();

		//
		// Now draw the plot components in the correct order.
		//
		drawNightTimeBackground();
		drawDayTimeBackground();
		drawMinorVerticalLines();
		drawMajorVerticalLines();
		drawWeekDayLabels();
		drawTimeLabels();
		drawHashedEdges();
		drawVerticalMarkerLines(mPlotInfo1);
		drawMinorHorizontalLines(mPlotInfo1);
		drawMajorHorizontalLines(mPlotInfo1);
		drawHorizontalMarkerLine(mPlotInfo1);
		drawHorizontalLabels(mPlotInfo1);
		drawSelectedValues(mPlotInfo2);
		drawAnnotations(mPlotInfo2);
		drawSelectedValues(mPlotInfo1);
		drawAnnotations(mPlotInfo1);

		mGraphics.setStroke(mDefaultStroke);

		//
		// Finally copy the finished image to the screen.
		//
		graphics.drawImage(mImage, 0, 0, null);
	}


	/**
	 * Calculate the minimum and maximum values to be plotted
	 * and the major and minor intervals for that range.
	 * This handles the case of either one plot or two plots.
	 * In the case of two plots, they can be related or unrelated.
	 */
	private void calculateValuesRange()
	{
		//
		// Calculate the ranges for the two plots assuming they are unrelated.
		//
		calculateValuesRange(mPlotInfo1);
		calculateValuesRange(mPlotInfo2);

		//
		// See whether there is a second plot and whether its values
		// are compatible with the first plot.  If not then do no more.
		//
		if (!mPlotInfo1.isValid() || !mPlotInfo2.isValid())
			return;

		if (mPlotInfo1.mSelector.getValueType() !=
			mPlotInfo2.mSelector.getValueType())
		{
			return;
		}

		if (!mPlotInfo2.mSelector.hasValues(mPlotInfo2.mFullObservationList))
			return;

		//
		// The second plot is compatible, so calculate new minimum
		// and maximum values to cover both of their ranges so that they
		// can both fit and use the same grid and labels.
		//
		double minimumValue =
			Math.min(mPlotInfo1.mMinimumValue, mPlotInfo2.mMinimumValue);

		double maximumValue =
			Math.max(mPlotInfo1.mMaximumValue, mPlotInfo2.mMaximumValue);

		//
		// Get the range for both plots and determine the major and minor
		// intervals for the range.  Only the first selector is used since it is the
		// primary one, and any second selector should be compatible.
		//
		double range = maximumValue - minimumValue;

		double majorInterval = mPlotInfo1.mSelector.getMajorInterval(range);
		double minorInterval = mPlotInfo1.mSelector.getMinorInterval(range);

		//
		// Round the minimum and maximum values to the major intervals.
		//
		maximumValue = Utils.roundUpToExactMultiple(maximumValue, majorInterval);
		minimumValue = Utils.roundDownToExactMultiple(minimumValue, majorInterval);

		//
		// Set the values into both plot information instances so that
		// they share the same values.
		//
		mPlotInfo1.mMinimumValue = minimumValue;
		mPlotInfo1.mMaximumValue = maximumValue;
		mPlotInfo1.mMajorInterval = majorInterval;
		mPlotInfo1.mMinorInterval = minorInterval;

		mPlotInfo2.mMinimumValue = minimumValue;
		mPlotInfo2.mMaximumValue = maximumValue;
		mPlotInfo2.mMajorInterval = majorInterval;
		mPlotInfo2.mMinorInterval = minorInterval;
	}


	/**
	 * Calculate the minimum and maximum values to be plotted
	 * and the major and minor intervals for that range.
	 * The range can be calculated either for the observations currently
	 * being shown in the plot, or for all observations that could be
	 * plotted by shifting.
	 */
	private void calculateValuesRange(PlotInfo plotInfo)
	{
		if (!plotInfo.isValid())
			return;

		//
		// Get the observation list to use for calculating the minimum
		// and maximum values depending on whether uniform
		// scaling is enabled.
		//
		ObservationList observationList = plotInfo.mObservationList;

		if (mPlotParameters.mIsUniformScaling)
			observationList = plotInfo.mFullObservationList;

		//
		// Get the minimum and maximum values for the plot.
		//
		double minimumValue = plotInfo.mSelector.getMinimumValue(observationList);
		double maximumValue = plotInfo.mSelector.getMaximumValue(observationList);

		//
		// Get the range and determine the major and minor
		// intervals for the range.
		//
		double range = maximumValue - minimumValue;

		plotInfo.mMajorInterval = plotInfo.mSelector.getMajorInterval(range);
		plotInfo.mMinorInterval = plotInfo.mSelector.getMinorInterval(range);

		//
		// Round the minimum and maximum values to the major intervals.
		//
		plotInfo.mMaximumValue = Utils.roundUpToExactMultiple(
			maximumValue, plotInfo.mMajorInterval);

		plotInfo.mMinimumValue = Utils.roundDownToExactMultiple(
			minimumValue, plotInfo.mMajorInterval);

		//
		// If the range is empty then make it at least one major interval high.
		//
		if (plotInfo.mMaximumValue <= plotInfo.mMinimumValue)
		{
			plotInfo.mMaximumValue =
				plotInfo.mMinimumValue + plotInfo.mMajorInterval;
		}
	}


	/**
	 * Draw the minor vertical lines for the plot.
	 * These are dashed.
	 */
	private void drawMinorVerticalLines()
	{
		mGraphics.setColor(mMinorVerticalLineColor);
		mGraphics.setStroke(mDashesStroke);

		for (int index = 0; index < mTimeTable.length; index++)
		{
			DateTime dateTime = mTimeTable[index];

			if (!isMinorInterval(dateTime))
				continue;

			int xPos = mSideInset + (mGridWidth * index) / (mTimeTable.length - 1);

			mGraphics.drawLine(xPos, mTopInset, xPos, mTopInset + mGridHeight);	
		}

		mGraphics.setStroke(mDefaultStroke);
	}


	/**
	 * Draw the major vertical lines for the plot for the major intervals.
	 */
	private void drawMajorVerticalLines()
	{
		mGraphics.setColor(mMajorVerticalLineColor);

		for (int index = 0; index < mTimeTable.length; index++)
		{
			DateTime dateTime = mTimeTable[index];

			if (!isMajorInterval(dateTime))
				continue;

			int xPos = mSideInset + (mGridWidth * index) / (mTimeTable.length - 1);

			//
			// Draw a normal or a wide line depending on the time of day and on
			// whether the time interval is large enough for that to matter.
			//
			mGraphics.drawLine(xPos, mTopInset, xPos, mTopInset + mGridHeight);

			if (dateTime.isExactDay() && mTimeMajorInterval < 24 * 60)
			{
				mGraphics.drawLine(xPos - 1, mTopInset, xPos - 1, mTopInset + mGridHeight);
				mGraphics.drawLine(xPos + 1, mTopInset, xPos + 1, mTopInset + mGridHeight);
			}
		}
	}


	/**
	 * Draw a horizontal marker line if enabled.
	 * A horizontal line is drawn for the marker observation's value if any.
	 * @param plotInfo
	 */
	private void drawHorizontalMarkerLine(PlotInfo plotInfo)
	{
		//
		// Get the observation if any that the marking is for.
		// If there isn't one then do nothing.
		//
		Observation markerObservation = mPlotParameters.mMarkerObservation;

		if (markerObservation == null)
			return;

		//
		// If the observation doesn't have a double value then do nothing.
		//
		Selector selector = plotInfo.mSelector;

		if (!selector.hasDoubleValue(markerObservation))
			return;

		//
		// If the observation's value is off of the plot then do nothing.
		// This can happen if an observation had been previously selected
		// from a different site than the site currently being displayed.
		//
		double value = selector.getDoubleValue(markerObservation);

		if ((value < mPlotInfo1.mMinimumValue) ||
			(value > mPlotInfo1.mMaximumValue))
		{
			return;
		}

		//
		// The value is valid so draw the horizontal line for it.
		//
		mGraphics.setColor(mMarkerLineColor);

		int yPos = getYCoordinate(plotInfo, value);

		mGraphics.drawLine(mSideInset, yPos, mSideInset + mGridWidth, yPos);
	}


	/**
	 * Draw the vertical marker lines if enabled.
	 * These are based on the specified marker observation if any.
	 * Multiple vertical lines are drawn in all intervals before and after the
	 * marker observation's UTC time (where the interval is a day or a year).
	 */
	private void drawVerticalMarkerLines(PlotInfo plotInfo)
	{
		//
		// Determine the vertical interval as either a day or a year,
		// or nothing.  If nothing, then draw no lines.
		//
		long interval = 0;

		if (mPlotParameters.mTimeInterval <= DateTime.DAY_INTERVAL * 60)
			interval = DateTime.DAY_INTERVAL;

		if (mPlotParameters.mTimeInterval >= DateTime.YEAR_INTERVAL)
			interval = DateTime.YEAR_INTERVAL;

		if (interval == 0)
			return;

		//
		// Get the observation if any that the marking is for.
		// If there isn't one then do nothing.
		//
		Observation markerObservation = mPlotParameters.mMarkerObservation;

		if (markerObservation == null)
			return;

		mGraphics.setColor(mMarkerLineColor);

		//
		// Draw a vertical line for each interval forwards from the marker's time.
		//
		DateTime dayTime = markerObservation.getUTCTime();

		while (dayTime.compareTo(mMaximumTime) <= 0)
		{
			if (dayTime.compareTo(mMinimumTime) >= 0)
			{
				int xPos = getXCoordinate(dayTime);

				mGraphics.drawLine(xPos, mTopInset, xPos, mTopInset + mGridHeight);
			}

			dayTime = dayTime.getAdjustedTime(interval);
		}

		//
		// Draw a vertical line for each interval backwards from the marker's time.
		//
		dayTime = markerObservation.getUTCTime();

		while (dayTime.compareTo(mMinimumTime) >= 0)
		{
			if (dayTime.compareTo(mMaximumTime) <= 0)
			{
				int xPos = getXCoordinate(dayTime);

				mGraphics.drawLine(xPos, mTopInset, xPos, mTopInset + mGridHeight);
			}

			dayTime = dayTime.getAdjustedTime(-interval);
		}
	}


	/**
	 * Draw the time labels at the bottom of the plot for the major intervals.
	 * This is only the date and hour values, not the weekdays.
	 */
	private void drawTimeLabels()
	{
		mGraphics.setColor(mLabelColor);
		mGraphics.rotate(Math.PI / 2);

		for (int index = 0; index < mTimeTable.length; index++)
		{
			DateTime dateTime = mTimeTable[index];

			if (!isMajorInterval(dateTime))
				continue;

			int xPos = mSideInset + (mGridWidth * index) / (mTimeTable.length - 1);

			//
			// For a day boundary give the month and year, otherwise give the hour
			// in the current day.
			//
			String label;

			if (dateTime.isExactDay())
			{
				label = String.format("%02d/%02d",
					dateTime.getMonth(), dateTime.getDay());
			}
			else
				label = String.format("%02d", dateTime.getHour());

			mGraphics.drawString(label, mTopInset + mGridHeight + 8, -xPos + 4);
		}

		mGraphics.setTransform(mDefaultTransform);
	}


	/**
	 * Draw the weekday names at the bottom of the plot.
	 * These are drawn under the 12:00 noon times in their
	 * own color so that it might not be distracting.
	 */
	public void drawWeekDayLabels()
	{
		//
		// Don't draw the week days at all if there isn't room.
		//
		int minutes = mMinimumTime.getMinutesTo(mMaximumTime);

		if (minutes > 60 * 24 * 90)
			return;

		//
		// Set to draw three characters or just one character
		// depending on the room.
		//
		int charCount = 3;
		int charOffset = 10;

		if (minutes > 60 * 24 * 30)
		{
			charCount = 1;
			charOffset = 3;
		}

		mGraphics.setColor(mDayOfWeekColor);

		for (int index = 0; index < mTimeTable.length; index++)
		{
			DateTime dateTime = mTimeTable[index];

			if (!isWeekDayInterval(dateTime))
				continue;

			int xPos = mSideInset + (mGridWidth * index) / (mTimeTable.length - 1);

			String label = dateTime.getDayOfWeekText().substring(0, charCount);

			mGraphics.drawString(label, xPos - charOffset,
				mTopInset + mGridHeight + mBottomInset - 8);
		}
	}


	/**
	 * Draw the minor horizontal lines for the grid.
	 * These lines are dashed.
	 */
	private void drawMinorHorizontalLines(PlotInfo plotInfo)
	{
		mGraphics.setColor(mMinorHorizontalLineColor);
		mGraphics.setStroke(mDashesStroke);

		double range = plotInfo.mMaximumValue - plotInfo.mMinimumValue;
		double verticalInterval =range / plotInfo.mMinorInterval;

		int verticalDivisions = (int) Math.round(Math.abs(verticalInterval));

		for (int i = 0; i <= verticalDivisions; i++)
		{
			int yPos = mTopInset + (mGridHeight * i ) / verticalDivisions;

			mGraphics.drawLine(mSideInset, yPos, mSideInset + mGridWidth, yPos);
		}

		mGraphics.setStroke(mDefaultStroke);
	}


	/**
	 * Draw the major horizontal lines for the grid.
	 */
	private void drawMajorHorizontalLines(PlotInfo plotInfo)
	{
		double range = plotInfo.mMaximumValue - plotInfo.mMinimumValue;
		double verticalInterval = range / plotInfo.mMajorInterval;
		int verticalDivisions = (int) Math.round(Math.abs(verticalInterval));

		mGraphics.setColor(mMajorHorizontalLineColor);

		for (int i = 0; i <= verticalDivisions; i++)
		{
			int yPos = mTopInset + (mGridHeight * i ) / verticalDivisions;
			mGraphics.drawLine(mSideInset, yPos, mSideInset + mGridWidth, yPos);

			//
			// If this line is to be bolded then do so.
			//
			double value = plotInfo.mMaximumValue - (range * i) / verticalDivisions;

			if (mPlotInfo1.mSelector.isValueBolded(value))
			{
				mGraphics.drawLine(mSideInset, yPos - 1, mSideInset + mGridWidth, yPos - 1);
				mGraphics.drawLine(mSideInset, yPos + 1, mSideInset + mGridWidth, yPos + 1);
			}
		}
	}


	/**
	 * Draw the labels for the horizontal grid lines on the left and right sides.
	 * These can be shown for just the major intervals or for both the major
	 * and minor intervals depending on the selector's choice.
	 */
	private void drawHorizontalLabels(PlotInfo plotInfo)
	{
		double range = plotInfo.mMaximumValue - plotInfo.mMinimumValue;

		double plotInterval = (plotInfo.mSelector.isMinorIntervalLabeled(range) ?
			plotInfo.mMinorInterval : plotInfo.mMajorInterval);

		double verticalInterval = range / plotInterval;
		int verticalDivisions = (int) Math.round(Math.abs(verticalInterval));

		mGraphics.setColor(mLabelColor);

		for (int i = 0; i <= verticalDivisions; i++)
		{
			int yPos = mTopInset + (mGridHeight * i ) / verticalDivisions;

			//
			// Format and draw the labels at the left and right sides.
			// Suppress the printing of a negative zero.
			//
			double value = plotInfo.mMaximumValue - (range * i) / verticalDivisions;

			String label = String.format("%.1f", value);

			if (label.equals("-0.0"))
				label = "0.0";

			mGraphics.drawString(label, 4, yPos + 4);
			mGraphics.drawString(label, mSideInset + mGridWidth + 6, yPos + 6);
		}
	}


	/**
	 * If necessary draw hashed left or right edges for the grid to indicate
	 * that shifting may be done to see more of the plot in that direction.
	 * This is based on whether the plot shows the first and last observations
	 * and on whether shifting is allowed.
	 */
	private void drawHashedEdges()
	{
		if ((mPlotParameters.mTimeInterval <= 0) ||
			(mPlotParameters.mTimeInterval == Long.MAX_VALUE))
		{
			return;
		}

		Observation observation = mPlotInfo1.mFullObservationList.getFirstObservation();

		if (mMinimumTime.compareTo(observation.getUTCTime()) > 0)
			drawHashedLine(mSideInset);

		observation = mPlotInfo1.mFullObservationList.getLastObservation();

		if (mMaximumTime.compareTo(observation.getUTCTime()) < 0)
			drawHashedLine(mSideInset + mGridWidth);
	}


	/**
	 * Draw hashes on the line at the specified X coordinate running from the
	 * bottom to the the top of the grid.  The hashes are short diagonal lines
	 * crossing the vertical line.
	 * @param XPosition
	 */
	private void drawHashedLine(int x)
	{
		mGraphics.setColor(mMajorVerticalLineColor);

		for (int y = mHashedLineInterval / 2;
			y < mGridHeight - mHashedLineInterval / 2;
			y += mHashedLineInterval)
		{
			mGraphics.drawLine(x - mHashedLineSize,
				y + mTopInset + mHashedLineSize,
				x + mHashedLineSize,
				y + mTopInset - mHashedLineSize);
		}
	}


	/**
	 * Draw the night time background.
	 * This is done before drawing the daytime so we can just fill the
	 * whole grid with the night time color.
	 */
	private void drawNightTimeBackground()
	{
		mGraphics.setColor(mNightTimeBackgroundColor);

		mGraphics.fillRect(mSideInset, mTopInset, mGridWidth, mGridHeight);
	}


	/**
	 * Calculate the sunrise and sunset times for the current observation list.
	 * The results are saved into the mSunRise and mSunSet arrays.
	 * This is only done if all the observations are consistent (i.e, the same site)
	 * the number of days being plotted isn't too large, and we aren't doing am
	 * interval plot.
	 */
	private void calculateSunriseSunsetTimes()
	{
		//
		// Clear the table until we know we want it.
		//
		mSunriseSunsetTable = null;

		//
		// First check that we want the sunrise and sunset times calculated.
		//
		if (!mIsConsistent || (mDayTimeDays <= 0) || (mPlotParameters.mGroupHours > 0))
			return;

		Observation firstObservation = mPlotInfo1.mFullObservationList.getFirstObservation();

		if (firstObservation == null)
			return;

		//
		// See if the number of days being plotted is larger than the
		// configured maximum value.  If so then we don't show daytime.
		//
		DateTime minNoonTime =mMinimumTime.roundToNoon();
		DateTime maxNoonTime = mMaximumTime.roundToNoon();

		int minutesSpanned = minNoonTime.getMinutesTo(maxNoonTime);

		if (minutesSpanned / (60 * 24) > mDayTimeDays)
			return;

		//
		// Calculate the number of "noons" which covers the time
		// range of the plot and allocate a table to hold data for each noon.
		//
		DateTime firstNoon = getNoonTime(mMinimumTime);
		DateTime lastNoon = getNoonTime(mMaximumTime);

		long firstNoonTime = firstNoon.getDateValue();
		long lastNoonTime = lastNoon.getDateValue();

		long noonDifference = lastNoonTime - firstNoonTime;

		int noonCount = ((int) Math.ceil(((double) noonDifference) / DateTime.DAY_INTERVAL)) + 1;

		mSunriseSunsetTable = new SunriseSunsetInfo[noonCount];

		//
		// Now calculate the sunrise and sunset times which surround
		// each of the noon times and save the results into the array.
		//
		SunriseSunset sunriseSunset = new SunriseSunset();

		double latitude = firstObservation.getLatitude();
		double longitude = firstObservation.getLongitude();

		long noonTime = firstNoonTime;
		long dateBeginTime = firstNoonTime - (DateTime.DAY_INTERVAL / 2);
		long dateEndTime = firstNoonTime + (DateTime.DAY_INTERVAL / 2);

		for (int index = 0; index < noonCount; index++)
		{
			//
			// Get the sunrise and sunset time for the site (if any), supplying a
			// GMT date which is a local noon for the site.  Note: The returned
			// values will be null if the sunrise or sunset doesn't exist on the day.
			//
			Date time = new Date(noonTime);

			Date sunrise = sunriseSunset.getSunrise(latitude, longitude, time, mGMTTimezone);
			Date sunset = sunriseSunset.getSunset(latitude, longitude, time, mGMTTimezone);

			//
			// If the times exist then extract them, fixing off by one day calculation mistakes.
			// This isn't always successful for extreme latitudes, but I don't know how to fix it.
			//
			long sunriseTime = 0;
			long sunsetTime = 0;

			if (sunrise != null)
			{
				sunriseTime = sunrise.getTime();

				if (sunriseTime > noonTime)
					sunriseTime -= DateTime.DAY_INTERVAL;

				if (sunriseTime < dateBeginTime)
					sunriseTime = dateBeginTime;
			}

			if (sunset != null)
			{
				sunsetTime = sunset.getTime();

				if (sunsetTime < noonTime)
					sunsetTime += DateTime.DAY_INTERVAL;

				if (sunsetTime > dateEndTime)
					sunsetTime = dateEndTime;
			}

			//
			// Determine if the sun is visible at some time of the day.
			// If there are no sunrises or sunsets then the sun is either
			// up all day or down all day,and we have to check the latitude
			// and the time of the year to determine which.
			//
			boolean isSunVisible = ((sunrise != null) || (sunset != null));

			if (!isSunVisible)
			{
				DateTime thisNoon = new DateTime(noonTime);
				int month = thisNoon.getMonth();
				int day = thisNoon.getDay();

				boolean isSunSouth = false;

				//
				// The sun is south of the equator between
				// (approximately) September 21 and March 21.
				// This will only be wrong right near the poles.
				//
				if ((month < 3) || ((month == 3) && (day < 21)))
					isSunSouth = true;

				if ((month > 9) || ((month == 9) && (day > 21)))
					isSunSouth = true;

				//
				// The sun is visible if its hemisphere is the same
				// as the site.
				//
				isSunVisible = (isSunSouth == (latitude < 0));
			}

			//
			// Store the info into the table and advance to the next noon time.
			//
			SunriseSunsetInfo info = new SunriseSunsetInfo(
				noonTime, isSunVisible, sunriseTime, sunsetTime);

			mSunriseSunsetTable[index] = info;

			noonTime += DateTime.DAY_INTERVAL;
			dateBeginTime +=  DateTime.DAY_INTERVAL;
			dateEndTime +=  DateTime.DAY_INTERVAL;
		}
	}


	/**
	 * Color the day time in the background using a lighter color than the night time.
	 * The night time background has already been drawn.
	 * This is only done if all the observations are consistent (i.e, the same site)
	 * and if the number of days being plotted isn't too large.
	 * This uses the previously calculated sunrise and sunset times.
	 */
	private void drawDayTimeBackground()
	{
		//
		// If the table wasn't created for any reason then do nothing.
		//
		if (mSunriseSunsetTable == null)
			return;

		mGraphics.setColor(mDayTimeBackgroundColor);

		//
		// Iterate over the table for each date to be plotted.
		// Each entry contains GMT times related to the date
		// specified by a local noon time at the site.
		//
		for (SunriseSunsetInfo info : mSunriseSunsetTable)
		{
			long sunrise = info.mSunrise;
			long sunset = info.mSunset;

			//
			// If the sun isn't visible on this day then skip it.
			//
			if (!info.mIsSunVisible)
				continue;

			//
			// If there isn't a sunrise or a sunset then pretend there
			// was one at the beginning or end of the day.
			//
			if (sunrise == 0)
				sunrise = info.mNoon - (DateTime.DAY_INTERVAL / 2);

			if (sunset == 0)
				sunset = info.mNoon + (DateTime.DAY_INTERVAL / 2);

			//
			// Truncate the times to the displayed interval and
			// then see if the interval is empty.
			//
			DateTime minTime = new DateTime(sunrise);
			DateTime maxTime = new DateTime(sunset);

			if (minTime.compareTo(mMinimumTime) < 0)
				minTime = mMinimumTime;

			if (maxTime.compareTo(mMaximumTime) > 0)
				maxTime = mMaximumTime;

			if (minTime.compareTo(maxTime) >= 0)
				continue;

			//
			// There is an interval to display, so do it.
			//
			int fromX = getXCoordinate(minTime);
			int toX = getXCoordinate(maxTime);

			mGraphics.fillRect(fromX, mTopInset, toX - fromX + 1, mGridHeight);
		}
	}


	/**
	 * Calculate the time values for the plotting of the vertical lines
	 * and for the labels at the bottom.  These times can be either the
	 * local times or the UTC times depending on the configuration.
	 * The results are saved in an array of local times for the plot and
	 * is valid until the plot is rescaled or shifted, or the site is changed.
	 * The array is needed since local times depend on daylight saving
	 * time changes, and so the major and minor intervals aren't
	 * uniform (sigh).  The array is indexed by 1/2 hour increments since
	 * some daylight savings times are off by that much.
	 */
	private void calculateTimeValues()
	{
		int totalMinutes = mMinimumTime.getMinutesTo(mMaximumTime);

		int intervalCount = (totalMinutes / 30) + 1;

		mTimeTable = new DateTime[intervalCount];

		for (int index = 0; index < intervalCount; index++)
		{
			DateTime dateTime = mMaximumTime.getAdjustedTime(-index * 30L * DateTime.MINUTE_INTERVAL);

			if (!mPlotParameters.mIsShowUTCTimes)
				dateTime = getLocalTime(dateTime);

			mTimeTable[intervalCount - index - 1] = dateTime;
		}
	}


	/**
	 * Calculate the GMT time of the local noon from the specified GMT time.
	 * @param time
	 * @return
	 */
	private DateTime getNoonTime(DateTime GMTTime)
	{
		DateTime localTime = getLocalTime(GMTTime);

		long GMTToLocalOffset = GMTTime.getOffsetTo(localTime);

		localTime = localTime.roundToNoon();

		return localTime.getAdjustedTime(-GMTToLocalOffset);
	}


	/**
	 * Determine whether or not the specified date time is a multiple
	 * of the major time interval.  This check splits into two cases
	 * according to whether the time interval is one day or larger.
	 * @param dateTime
	 * @return
	 */
	private boolean isMajorInterval(DateTime dateTime)
	{
		//
		// Split the major time interval into days and parts
		// of days so we can handle them separately.
		//
		int dayInterval = mTimeMajorInterval / (60 * 24);
		int minuteInterval = mTimeMajorInterval % (60 * 24);

		//
		// If the interval is a multiple of days then check that
		// the time is an exact day and a multiple of the interval.
		// Note: For this purpose the day of the year is used which
		// can cause small hiccups across year boundaries.
		//
		if (dayInterval >0)
		{
			if (!dateTime.isExactDay())
				return false;

			int dayOfYear = dateTime.getDayOfYear() - 1;

			if ((dayOfYear % dayInterval) != 0)
				return false;

			return true;
		}

		//
		// If the interval is a number of minutes then check that
		// the time of day is an exact multiple of that.
		//
		if (minuteInterval > 0)
		{
			int minuteOfDay = dateTime.getMinutesIntoDay();

			if ((minuteOfDay % minuteInterval) != 0)
				return false;
		}

		return true;
	}


	/**
	 * Determine whether or not the specified date time is one
	 * used for drawing the week days.  This check splits into two
	 * cases according to whether the time interval is one day or larger.
	 * @param dateTime
	 * @return
	 */
	private boolean isWeekDayInterval(DateTime dateTime)
	{
		if (!dateTime.isExactHour() || (dateTime.getHour() != 12))
			return false;

		//
		// Split the major time interval into days and parts
		// of days so we can handle them separately.
		//
		int dayInterval = mTimeMajorInterval / (60 * 24);
//		int minuteInterval = mTimeMajorInterval % (60 * 24);

		//
		// If the interval is a multiple of days then check that the
		// time is an exact noontime and a multiple of the interval.
		// Note: For this purpose the day of the year is used which
		// can cause small hiccups across year boundaries.
		//
		if (dayInterval >0)
		{
			if (!dateTime.isExactHour() || (dateTime.getHour() != 12))
				return false;

			int dayOfYear = dateTime.getDayOfYear() - 1;

			if ((dayOfYear % dayInterval) != 0)
				return false;

			return true;
		}

		//
		// The minute check below is not used since the half-day
		// check above seems good enough.
		//
		return true;
		//
		// If the interval is a number of minutes then check that
		// the time of day is an exact multiple of that but on the
		// middle of the major interval.
		//
//		int halfMinuteInterval = minuteInterval / 2;
//
//		if (halfMinuteInterval > 0)
//		{
//			int minuteOfDay = dateTime.getMinutesIntoDay();
//
//			if ((minuteOfDay % minuteInterval) == 0)
//				return false;
//
//			if ((minuteOfDay % halfMinuteInterval) != 0)
//				return false;
//		}
//
//		return true;
	}


	/**
	 * Determine whether or not the specified date time is a multiple
	 * of the minor time interval.  This check splits into two cases
	 * according to whether the time interval is one day or larger.
	 * In addition, a minor interval cannot be a major interval.
	 * @param dateTime
	 * @return
	 */
	private boolean isMinorInterval(DateTime dateTime)
	{
		if (isMajorInterval(dateTime))
			return false;

		//
		// Split the minor time interval into days and parts
		// of days so we can handle them separately.
		//
		int dayInterval = mTimeMinorInterval / (60 * 24);
		int minuteInterval = mTimeMinorInterval % (60 * 24);

		//
		// If the interval is a multiple of days then check that
		// the time is an exact day and a multiple of the interval.
		// Note: For this purpose the day of the year is used which
		// can cause small hiccups across year boundaries.
		//
		if (dayInterval >0)
		{
			if (!dateTime.isExactDay())
				return false;

			int dayOfYear = dateTime.getDayOfYear() - 1;

			if ((dayOfYear % dayInterval) != 0)
				return false;

			return true;
		}

		//
		// If the interval is a number of minutes then check that
		// the time of day is an exact multiple of that.
		//
		if (minuteInterval > 0)
		{
			int minuteOfDay = dateTime.getMinutesIntoDay();

			if ((minuteOfDay % minuteInterval) != 0)
				return false;
		}

		return true;
	}


	/**
	 * Return the Local time corresponding to the specified GMT time value.
	 * This uses the time offset calculated from the closest Observation to
	 * the GMT time.  This is close to correct, except possibly for times near
	 * the exact time of changes between standard and daylight savings.
	 * (We do this instead of trying to use a TimeZone because the timezone
	 * names in the collected reports are unusable.)
	 * @param GMTTime
	 */
	private DateTime getLocalTime(DateTime GMTTime)
	{
		Observation closestObservation = mPlotInfo1.mFullObservationList.getClosestObservation(GMTTime);

		if (closestObservation == null)
			return GMTTime;

		long GMTToLocalOffset = closestObservation.getLocalTimeOffset();

		DateTime localTime = GMTTime.getAdjustedTime(GMTToLocalOffset);

		return localTime;
	}


	/**
	 * Draw the observation values for the specified plot.
	 * The data points are drawn as circles with lines connecting them
	 * if successive observations are considered connected.
	 */
	private void drawSelectedValues(PlotInfo plotInfo)
	{
		if (!plotInfo.isValid())
			return;

		Selector selector = plotInfo.mSelector;

		DateTime prevTime = null;
		double prevValue = 0;
		Observation prevObservation = null;

		for (Observation observation : plotInfo.mObservationList)
		{
			if (!selector.hasDoubleValue(observation))
				continue;

			DateTime time = observation.getUTCTime();
			double value = selector.getDoubleValue(observation);

			drawPoint(plotInfo, time, value,
				(observation == mPlotParameters.mSelectedObservation));

			if (selector.isJoined(mPlotParameters, prevObservation, observation))
				drawLine(plotInfo, prevTime, prevValue, time, value);

			prevTime = time;
			prevValue = value;
			prevObservation = observation;
		}
	}


	/**
	 * Draw a circle representing an observation point.
	 * The circle's radius is reduced to just 1 pixel if the number of days
	 * shown in the plot is large enough.  The circle can be drawn open
	 * or filled.
	 */
	private void drawPoint(PlotInfo plotInfo, DateTime time,
		double value, boolean isFilled)
	{
		int x = getXCoordinate(time);
		int y = getYCoordinate(plotInfo, value);

		int radius = mPointRadius;

		if ((mPointShrinkageDays > 0) &&
			(mMinimumTime.getMinutesTo(mMaximumTime) >= mPointShrinkageDays * 60 * 24))
		{
			radius = 1;
		}

		mGraphics.setColor(plotInfo.mPointColor);

		if (isFilled)
		{
			mGraphics.fillArc(x - radius, y - radius, radius * 2, radius * 2, 0, 360);
		}
		else
		{
			mGraphics.drawArc(x - radius, y - radius, radius * 2, radius * 2, 0, 360);
		}
	}


	private void drawLine(PlotInfo plotInfo, DateTime time1, double value1,
		DateTime time2, double value2)
	{
		int x1 = getXCoordinate(time1);
		int x2 = getXCoordinate(time2);
		int y1 = getYCoordinate(plotInfo, value1);
		int y2 = getYCoordinate(plotInfo, value2);

		mGraphics.setColor(plotInfo.mLineColor);
		mGraphics.drawLine(x1, y1, x2, y2);
	}


	/**
	 * Draw annotations for the observations.
	 * This is a special-case method.
	 * The only annotations are to show wind direction
	 * vectors for the wind speed plot.
	 */
	private void drawAnnotations(PlotInfo plotInfo)
	{
		if (!plotInfo.isValid())
			return;

		//
		// Annotations are only applied to the wind speed
		// and only if the wind direction radius is nonzero.
		//
		if (plotInfo.mSelector != Selector.WIND_SPEED_KMH_SELECTOR)
			return;

		if (mWindDirectionRadius <= 0)
			return;

		mGraphics.setColor(plotInfo.mAnnotationColor);

		for (Observation observation : plotInfo.mObservationList)
		{
			//
			// If there isn't any wind speed then ignore
			// this observation.
			//
			if (!plotInfo.mSelector.hasDoubleValue(observation))
				continue;

			//
			// Get the bearing of the wind direction.
			// If there isn't any then ignore this observation.
			//
			Double bearing = Selector.WIND_DIRECTION_SELECTOR.
				getBearing(observation);

			if (bearing == null)
				continue;

			//
			// Calculate the center of the wind speed point.
			//
			DateTime time = observation.getUTCTime();
			double value = plotInfo.mSelector.getDoubleValue(observation);

			int x = getXCoordinate(time);
			int y = getYCoordinate(plotInfo, value);

			//
			// Calculate the vector to draw based on the
			// bearing and draw a line from the center of
			// the circle outwards.
			//
			double angle = Math.toRadians(90.0 - bearing);

			int rx = (int) Math.round(Math.cos(angle) * mWindDirectionRadius);
			int ry = (int) Math.round(Math.sin(angle) * mWindDirectionRadius);

			mGraphics.drawLine(x, y, x + rx, y - ry);
		}
	}


	/**
	 * Get the X coordinate for drawing in the panel based on the given date value.
	 * This is usually a value within the displayed grid.
	 * @param time
	 * @return
	 */
	private int getXCoordinate(DateTime time)
	{
		int range = mMinimumTime.getMinutesTo(mMaximumTime);
		int used = mMinimumTime.getMinutesTo(time);

		double scaledValue = ((double) used) / range;

		int coordinate = ((int) Math.round(scaledValue * mGridWidth)) + mSideInset;

		return coordinate;
	}


	/**
	 * Get the Y coordinate for drawing in the panel based on the given value.
	 * This is usually a value within the displayed grid.
	 */
	private int getYCoordinate(PlotInfo plotInfo, double value)
	{
		double range = plotInfo.mMaximumValue - plotInfo.mMinimumValue;
		double used = value - plotInfo.mMinimumValue;

		double scaledValue = used / range;

		int coordinate = mTopInset + mGridHeight - ((int) Math.round(scaledValue * mGridHeight));

		return coordinate;
	}


	/**
	 * Set the Color values as defined in the configuration file.
	 * Default colors are used in case of missing or illegal values.
	 */
	private void getColors()
	{
		Configuration config = Configuration.getInstance();

		//
		// The order of obtaining these matters a bit since some
		// default colors use the previously obtained colors.
		//
		mBackgroundColor = GUIUtils.getColorFromName(
			config.getBackgroundColor(),
			Color.blue.darker().darker().darker().darker());

		mDayTimeBackgroundColor = GUIUtils.getColorFromName(
			config.getDayTimeBackgroundColor(),
			Color.blue.darker());

		mNightTimeBackgroundColor = GUIUtils.getColorFromName(
			config.getNightTimeBackgroundColor(),
			mDayTimeBackgroundColor.darker());

		mMajorHorizontalLineColor = GUIUtils.getColorFromName(
			config.getMajorHorizontalLineColor(),
			Color.green.darker());

		mMinorHorizontalLineColor = GUIUtils.getColorFromName(
			config.getMinorHorizontalLineColor(),
			mMajorHorizontalLineColor.darker());

		mMajorVerticalLineColor = GUIUtils.getColorFromName(
			config.getMajorVerticalLineColor(),
			Color.green.darker());

		mMinorVerticalLineColor = GUIUtils.getColorFromName(
			config.getMinorVerticalLineColor(),
			mMajorVerticalLineColor.darker());

		mMarkerLineColor = GUIUtils.getColorFromName(
			config.getMarkerLineColor(),
			Color.red.darker());

		mPlotInfo1.mPointColor = GUIUtils.getColorFromName(
			config.getPointColor(),
			Color.yellow);

		mPlotInfo1.mLineColor = GUIUtils.getColorFromName(
			config.getLineColor(),
			mPlotInfo1.mPointColor);

		mPlotInfo1.mAnnotationColor = GUIUtils.getColorFromName(
			config.getAnnotationColor(),
			Color.red);

		mPlotInfo2.mPointColor = GUIUtils.getColorFromName(
			config.getPointColor2(),
			mPlotInfo1.mPointColor.darker().darker().darker());

		mPlotInfo2.mLineColor = GUIUtils.getColorFromName(
			config.getLineColor2(),
			mPlotInfo1.mLineColor.darker().darker().darker());

		mPlotInfo2.mAnnotationColor = GUIUtils.getColorFromName(
			config.getAnnotationColor2(),
			mPlotInfo1.mAnnotationColor.darker().darker());

		mLabelColor = GUIUtils.getColorFromName(
			config.getLabelColor(),
			Color.yellow);

		mDayOfWeekColor = GUIUtils.getColorFromName(
			config.getDayOfWeekColor(),
			Color.yellow.darker().darker().darker());
	}


	/**
	 * Method called on a key typed event.
	 * @param event
	 */
	protected void handleKeyTyped(KeyEvent event)
	{
		char ch = event.getKeyChar();
		int modifiers = event.getModifiersEx();

		boolean isControlDown =
			((modifiers & KeyEvent.CTRL_DOWN_MASK) != 0);

		switch (ch)
		{
			case 'q':
				mController.exit();
				return;

			case 'u':
				mController.updateReport();
				return;

			default:
				break;
		}

		if ((ch >= '0') && (ch <= '9'))
		{
			if (isControlDown)
				mController.addSelector(ch - '0');
			else
				mController.setSelector(ch - '0');

			return;
		}

		//
		// Look for the shifted digits as an alternative to using
		// the control modifier since Windows doesn't give
		// control-digits to Java.
		//
		int index = ")!@#$%^&*(".indexOf(ch);

		if (index >= 0)
		{
			mController.addSelector(index);

			return;
		}

		//
		// If the character is a letter then set the report
		// using a site key of that letter.
		//
		if (Utils.isLetter(ch))
		{
			mController.setReport(String.valueOf(ch));
		}
	}


	/**
	 * Method called on a key pressed event.
	 * This is used to shift the view left or right by the major interval
	 * (or by days if shift/control is used).
	 */
	protected void handleKeyPressed(KeyEvent event)
	{
		int keyCode = event.getKeyCode();
		int modifiers = event.getModifiersEx();

		int shiftHours = mTimeMajorInterval / 60;

		if ((modifiers & (KeyEvent.CTRL_DOWN_MASK | KeyEvent.SHIFT_DOWN_MASK)) != 0)
			shiftHours = 24;

		switch (keyCode)
		{
			case KeyEvent.VK_LEFT:
				mController.shiftViewLeft(shiftHours);
				break;

			case KeyEvent.VK_RIGHT:
				mController.shiftViewLeft(-shiftHours);
				break;
		}
	}


	/**
	 * Method called on a mouse clicked event.
	 * @param event
	 */
	protected void handleMouseClicked(MouseEvent event)
	{
		int button = event.getButton();
		Point point = event.getPoint();

		if (button == 1)
		{
			handleLeftClick(point.x, point.y);

			return;
		}

		if (button == 3)
		{
			handleRightClick(point.x, point.y);

			return;
		}
	}


	/**
	 * Handle a left click event on at the specified coordinates.
	 * This is used to bring up a details dialog box for the selected
	 * observation or to shift the view.
	 */
	private void handleLeftClick(int targetX, int targetY)
	{
		//
		// Shift left if the labels at the left are clicked.
		//
		if (targetX < mSideInset)
		{
			int hours = mMinimumTime.getMinutesTo(mMaximumTime) / 60;
			mController.shiftViewLeft(hours);

			return;
		}

		//
		// Shift right if the labels at the right are clicked.
		//
		if (targetX > mSideInset + mGridWidth)
		{
			int hours = mMinimumTime.getMinutesTo(mMaximumTime) / 60;
			mController.shiftViewLeft(-(hours));

			return;
		}

		//
		// The click is within the grid so search for the nearest observation.
		// If one is found then tell the controller to select it.
		//
		Observation observation = findObservation(mPlotInfo1, targetX, targetY);
		boolean isPlot1 = true;

		if (observation == null)
		{
			observation = findObservation(mPlotInfo2, targetX, targetY);
			isPlot1 = false;
		}

		if (observation != null)
			mController.selectObservation(isPlot1, observation);
	}


	/**
	 * Find the best observation which is close to the specified coordinates.
	 * Returns null if the coordinates are not close to any observation.
	 */
	private Observation findObservation(PlotInfo plotInfo, int targetX, int targetY)
	{
		if (!plotInfo.isValid())
			return null;

		Observation bestObservation = null;
		int bestDistance = 0;

		for (Observation observation : plotInfo.mObservationList)
		{
			if (!plotInfo.mSelector.hasDoubleValue(observation))
				continue;

			int x = getXCoordinate(observation.getUTCTime());
			int y = getYCoordinate(plotInfo, plotInfo.mSelector.getDoubleValue(observation));

			int xDiff = x - targetX;
			int yDiff = y - targetY;

			int distance = xDiff * xDiff + yDiff * yDiff;

			if (distance > mPointRadius * mPointRadius)
				continue;

			if ((bestObservation == null) || (distance < bestDistance))
			{
				bestObservation = observation;
				bestDistance = distance;
			}
		}

		return bestObservation;
	}


	/**
	 * Handle a right click event at the specified coordinates.
	 * This is used to bring up an actions dialog.
	 */
	private void handleRightClick(int targetX, int targetY)
	{
		mController.showActionsDialog();
	}


	/**
	 * A convenience class to hold all of the related information for a single plot.
	 * This information is only used for the plotting of the points and lines for
	 * the observation values, and keeps no data about the labels or time. 
	 */
	private static class PlotInfo
	{
		public ObservationList mFullObservationList;
		public ObservationList mObservationList;
		public Selector mSelector;
		public Color mLineColor;
		public Color mPointColor;
		public Color mAnnotationColor;
		public double mMaximumValue;
		public double mMinimumValue;
		public double mMajorInterval;
		public double mMinorInterval;


		/**
		 * Return whether or not this plot is valid for use.
		 */
		public boolean isValid()
		{
			return ((mObservationList != null) && (mSelector != null));
		}
	}


	/**
	 * A convenience class to hold the calculated information about noon,
	 * sunrise and sunset times for a day at a particular site.
	 * All of these times are in GMT but as applied to the local site.
	 * The noon value is the time of local noon for the site, and the sunrise
	 * and sunset times are the ones for the same local day as the noon time.
	 * The sunrise or sunset times are zero if they do not occur on the day
	 * (e.g., the site is too far south).
	 */
	private static class SunriseSunsetInfo
	{
		public final long mNoon;
		public final boolean mIsSunVisible;
		public final long mSunrise;
		public final long mSunset;


		public SunriseSunsetInfo(long noon, boolean isSunVisible, long sunrise, long sunset)
		{
			mNoon = noon;
			mIsSunVisible = isSunVisible;
			mSunrise = sunrise;
			mSunset = sunset;
		}
	}
}
